'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var shiki = require('shiki');

function isObject(item) {
    return !!item && typeof item === 'object' && !Array.isArray(item);
}
function hasOwn(target, key) {
    return Object.prototype.hasOwnProperty.call(target, key);
}
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function merge(target, source) {
    const output = { ...target };
    if (isObject(target) && isObject(source)) {
        Object.keys(source).forEach((key) => {
            if (isObject(source[key])) {
                if (!hasOwn(target, key)) {
                    Object.assign(output, {
                        [key]: source[key]
                    });
                }
                else {
                    Object.assign(output[key], merge(target[key], source[key]));
                }
            }
            else {
                Object.assign(output, {
                    [key]: source[key]
                });
            }
        });
    }
    return output;
}
function groupBy(elements, getter) {
    const map = new Map();
    for (const element of elements) {
        const key = getter(element);
        if (map.has(key)) {
            const group = map.get(key);
            if (group) {
                group.push(element);
            }
        }
        else {
            map.set(key, [element]);
        }
    }
    return map;
}

function mergeOptions$1(defaults, user) {
    return merge(defaults, user);
}
function getDefaultOptions$1() {
    return {
        openToken: '{{',
        closeToken: '}}',
        withLanguage: false,
        withLineNumbers: false,
        highlight: [],
        highlightClasses: ['highlight'],
        highlightRangeClasses: ['highlight-start', 'highlight-inner', 'highlight-end'],
        highlightInvert: false,
        highlightInvertClasses: ['highlight-invert'],
        highlighter: {}
    };
}

function mergeOptions(defaults, user) {
    return merge(defaults, user);
}
function getDefaultOptions() {
    return {
        withLanguage: false,
        withLineNumbers: false,
        languageId: '',
        lineOptions: [],
        foreground: '#000',
        background: '#fff'
    };
}

/** Simple string builder. */
function useBuilder(initial = []) {
    const strings = Array.from(initial);
    const setterFn = (value) => strings.push(value);
    const getterFn = (separator = '') => strings.join(separator);
    return [getterFn, setterFn];
}

function renderToHtml(tokens, userOptions = {}) {
    const options = mergeOptions(getDefaultOptions(), userOptions);
    const [html, append] = useBuilder();
    // NOTE: This is a check for the edge case when we render tokens from a fenced code block without
    // language defined, i.e. plain text.
    const lines = isPlainText(tokens) ? transformPlainText(tokens) : tokens;
    // Process possible edge case when we have an array of options for the same line, but with
    // differrent class names.
    const groupedLineOptions = groupBy(options.lineOptions, ({ line }) => line);
    // Open: building.
    append(`<pre class="shiki" style="background-color: ${options.background}">`);
    // Optionally add a div with `language-{languageId}` class and the `languageId` label itself.
    if (options.withLanguage && options.languageId.length) {
        append(`<div class="language-id">${options.languageId}</div>`);
    }
    // Open: building the code section.
    append(`<code>`);
    // Optionally add a div with line numbers.
    if (options.withLineNumbers) {
        append(`<div class="line-numbers">
        ${lines.map((_, lineIndex) => `<div class="line-number">${lineIndex + 1}</div>`).join('')}
      </div>`);
    }
    // Open: building the lines with code.
    append(`<div class="lines">`);
    // console.log(lines)
    lines.forEach((line, lineIndex) => {
        const lineNumber = lineIndex + 1;
        const lineOptions = groupedLineOptions.get(lineNumber) ?? [];
        const lineClasses = getLineClasses(lineOptions).join(' ');
        // Open: line.
        append(`<div class="${lineClasses}">`);
        if (line.length) {
            line.forEach((token) => {
                const { color, fontStyle, content } = token;
                const cssDeclarations = [
                    `color: ${color ?? options.foreground}`,
                    ...(fontStyle && fontStyle & shiki.FontStyle.Italic ? ['font-style: italic'] : []),
                    ...(fontStyle && fontStyle & shiki.FontStyle.Bold ? ['font-weight: bold'] : []),
                    ...(fontStyle && fontStyle & shiki.FontStyle.Underline ? ['text-decoration: underline'] : [])
                ];
                append(`<div class="line-token" style="${cssDeclarations.join('; ')}">${escapeHtml(content)}</div>`);
            });
        }
        else {
            append(`<div class="line-token">&nbsp;</div>`);
        }
        // Close: line.
        append(`</div>`);
    });
    // Close: building the lines with code.
    append(`</div>`);
    // Close: building the code section.
    append(`</code>`);
    // Close: building.
    append(`</pre>`);
    return html();
}
function escapeHtml(html) {
    // prettier-ignore
    const htmlEscapes = {
        '&': '&amp;',
        '<': '&lt;',
        '>': '&gt;',
        '"': '&quot;',
        '\'': '&#39;'
    };
    return html.replace(/[&<>"']/g, (char) => htmlEscapes[char]);
}
function isPlainText(tokens) {
    const [line] = tokens ?? [[]];
    const [token] = line ?? [];
    const hasSingleLine = tokens.length === 1;
    const hasSingleToken = line && line.length === 1;
    const hasContentOnly = token && !token.color && !token.fontStyle && !!token.content;
    return hasSingleLine && hasSingleToken && hasContentOnly;
}
function transformPlainText(tokens) {
    const [line] = tokens;
    const [token] = line;
    const strings = token.content.split('\n');
    const lines = strings
        .map((content) => ({ content }))
        .map((token) => (token.content.length ? [token] : []));
    return lines;
}
function getLineClasses(lineOptions) {
    const lineClasses = new Set(['line']);
    for (const lineOption of lineOptions) {
        for (const lineClass of lineOption.classes ?? []) {
            lineClasses.add(lineClass);
        }
    }
    return Array.from(lineClasses);
}

async function shikigami(userOptions = {}) {
    const options = mergeOptions$1(getDefaultOptions$1(), userOptions);
    const highlighter = await shiki.getHighlighter(options.highlighter);
    return function plugin(md) {
        if (md.renderer.rules.fence) {
            md.renderer.rules.fence = (markdownTokens, markdownTokenIndex) => {
                const markdownToken = markdownTokens[markdownTokenIndex];
                const markdownContent = markdownToken.content.trim();
                const { meta, languageId } = extractMeta(markdownToken.info.trim(), {
                    openToken: options.openToken,
                    closeToken: options.closeToken
                });
                const withLanguage = options.withLanguage;
                const withLineNumbers = meta.withLineNumbers ?? options.withLineNumbers;
                const themedTokensRaw = highlighter.codeToThemedTokens(markdownContent, languageId);
                const themedTokens = filterOutTrailingEmptyTokens(themedTokensRaw);
                const lineOptions = createLineOptions(themedTokens.length, meta.highlight ?? options.highlight, {
                    classes: options.highlightClasses,
                    rangeClasses: options.highlightRangeClasses,
                    invertClasses: options.highlightInvert ? options.highlightInvertClasses : []
                });
                const foreground = highlighter.getForegroundColor();
                const background = highlighter.getBackgroundColor();
                return renderToHtml(themedTokens, {
                    withLanguage,
                    withLineNumbers,
                    languageId,
                    lineOptions,
                    foreground,
                    background
                });
            };
        }
    };
}
function filterOutTrailingEmptyTokens(sourceTokens) {
    const { tokens } = sourceTokens.reduceRight(({ tokens, hitNonEmpty }, next) => !hitNonEmpty
        ? next.length
            ? { hitNonEmpty: true, tokens: [next, ...tokens] }
            : { hitNonEmpty, tokens }
        : { hitNonEmpty, tokens: [next, ...tokens] }, { tokens: [], hitNonEmpty: false });
    return tokens;
}
function extractMeta(source, { openToken, closeToken }) {
    const hasOpenBraces = source.includes(openToken);
    const hasCloseBraces = source.includes(closeToken);
    switch (true) {
        case hasOpenBraces && hasCloseBraces: {
            const pos = {
                start: source.indexOf(openToken),
                end: source.lastIndexOf(closeToken)
            };
            const metaRaw = source.substring(pos.start + openToken.length, pos.end).trim();
            const languageId = source.substring(0, pos.start).trim();
            try {
                // NOTE: Bruteforce and low-effort solution.
                const resolveToMetaPojo = new Function(`return { ${metaRaw} }`);
                const metaPojo = resolveToMetaPojo();
                const meta = sanitizeMetaPojo(metaPojo);
                return {
                    meta,
                    languageId
                };
            }
            catch {
                return {
                    meta: {},
                    languageId
                };
            }
        }
        default: {
            const languageId = source.substring(0, source.length);
            return {
                meta: {},
                languageId
            };
        }
    }
}
function sanitizeMetaPojo(metaPojo) {
    const ALLOWED_KEYS = [
        'withLanguage',
        'withLineNumbers',
        'highlight',
        'highlightClasses',
        'highlightRangeClasses'
    ].map((key) => key.toLowerCase());
    return Object.entries(metaPojo)
        .filter(([key]) => ALLOWED_KEYS.includes(key.toLowerCase()))
        .reduce((acc, [key, value]) => ({ ...acc, [key]: value }), {});
}
function createLineOptions(sourceLinesAmount, highlightLines, { classes, rangeClasses, invertClasses }) {
    const highlightedLines = highlightLines.reduce((acc, next) => Array.isArray(next)
        ? [...acc, ...mapRange(next, rangeClasses)]
        : [...acc, ...mapFlat([next], classes)], []);
    const sourceLines = sourceLinesAmount && highlightedLines.length
        ? Array.from({ length: sourceLinesAmount }, (_, index) => ({
            line: index + 1,
            classes: invertClasses
        })).filter((sourceLine) => !highlightedLines.some((highlightedLine) => highlightedLine.line === sourceLine.line))
        : [];
    return [...sourceLines, ...highlightedLines];
}
function mapFlat(lines, classes) {
    return lines.map((line) => ({ line, classes }));
}
function mapRange([start, end], [startClass, innerClass, endClass]) {
    const length = Math.abs(end - start) + 1;
    const isAscending = start <= end;
    const lines = Array.from({ length }, (_, index) => (isAscending ? index + start : start - index));
    const rangeStart = lines.slice(0, 1);
    const rangeInner = lines.length > 2 ? lines.slice(1, -1) : [];
    const rangeEnd = lines.slice(-1);
    return [
        ...mapFlat(rangeStart, [startClass]),
        ...mapFlat(rangeInner, [innerClass]),
        ...mapFlat(rangeEnd, [endClass])
    ];
}

Object.defineProperty(exports, 'loadTheme', {
    enumerable: true,
    get: function () { return shiki.loadTheme; }
});
exports.shikigami = shikigami;
//# sourceMappingURL=index.cjs.map
